/*******************************************************************************

Although this program has been used by the USGS, no warranty, expressed or
implied, is made by the USGS or the United States Government as to the
accuracy and functioning of the program and related program material nor
shall the fact of distribution constitute any such warranty, and no
responsibility is assumed by the USGS in connection therewith.


(a) USGS Computer Program: NdmOls082.c

(b) Use of this program is described in USGS Report OFR 2012-[TBD]: "Computing
    Ordinary Least-Squares Parameter Estimates for the National Descriptive
    Model of Mercury in Fish" by D. Donato

(c) Program written in C. This is a preliminary version 0.82.

(d) Program has been successfully compiled using GCC 4.4.4 under Linux Kernel Version 2.6.33.

(e) Code for this program is released in source form.


 -- January 27, 2012 --


/*******************************************************************************

     PROGRAM:                              Fit Parameters to the NDMMF
                                           in the Ordinary Least Squares Sense

     PROGRAM FILE NAME:                    NdmOls082.c

     VERSION:                              0.82 

     SYSTEM:                               EMMMA/NDMMF

     AUTHOR:                               D. Donato

     DATE FIRST USED:                      April 16, 2011
                                           April 17, 2011

     MODIFICATION DATES:                   April 17, 2011
                                           June 23, 2011
                                           January 27, 2012

     DESCRIPTION:

     This program finds the best-fit parameters in the Ordinary Least Squares
     (OLS) sense for the National Descriptive Model of Mercury in Fish
     (NDMMF).

     This program solves the normal equations,

           ( X' X ) b  =  X' Y

     using the method of Gaussian Elimination (GE). Although it may be more
     common for the normal equations to be solved by inverting ( X' X) with
     the normal equations rewritten as

                        -1
          b  =  ( X' X )     X' Y

     GE is used instead because (i) it is a straightforward technique, and
     (ii) in the case of the NDMMF, both the matrix ( X' X) and the matrix
     X' Y can be prepared in a small amount of computational time during
     input of the file of observations without storing either matrix X
     or its transpose X' in memory. Separate documentation explains
     this in greater detail.

     Immediately after the observations have been input from the input data
     file, this program is ready to proceed with GE. The matrix ( X' X )
     is a p x p matrix, with p = the number of variables in the NDMMF.
     More specifically, p = the number of different SPC values + the number
     of events.
     
     The array variable y is used for the matrix product X'Y and the
     two-dimensional array variable p is used for the matrix product X'X.


     MODIFICATION DESCRIPTIONS:

     April 17, 2011 -- Version 0.80 has several defects:
          (i)    It does not use weighting information;
          (ii)   It performs lengthy computation during Gaussian
                 Elimination; and
          (iii)  It tests for a non-zero value on the diagonal of
                 the product matrix and offers to swap rows even
                 though there should be no non-zero values on the
                 diagaonal and a row swap would leave the association
                 between a parameter and its variable identification
                 in doubt.
        This version aims to correct two of these three defects.
        Defect (ii) remains in this version because a technique for
        reducing the volume of computation has not been found -- yet.

     June 23, 2011 -- Version 0.82 has been slightly modified to accommodate
        a floating-point weight and a recordid in each input filed. The
        recordid is input but not stored and not used.

     January 27, 2012 -- Minor annotations were added.
    

*******************************************************************************/

/*******************************************************************************
                             PRE-PROCESSOR DIRECTIVES
*******************************************************************************/

#include <stdlib.h>
#include <math.h>
#include <stdio.h>

#define  MATRIXDIMENSION  15550 
#define  NUMSPC             566
#define  NUMEVENTS        14922

/*******************************************************************************
                                  GLOBAL VARIABLES
*******************************************************************************/

  /*** Matrix and Column Vectors         ***/

  double  p[MATRIXDIMENSION][MATRIXDIMENSION];
  double  b[MATRIXDIMENSION], y[MATRIXDIMENSION];

  /*** Double-Precision Constants        ***/

  double zero={(double)0.0};
  double  one={(double)1.0};
  double onek={(double)1000.0};

  /*** Characteristics of the Data Case  ***/

  int numvariables;

  /*** Values to Be Accepted From Inputs ***/
  
  double     Resultin, Inchesin;
  int        SPCin, Eventin, DLin, WTin;

  /*** Values to Be Computed From Inputs ***/
  
  double      L, Y;

  /*** File-Related Variables ***/

  FILE  *IN1, *OUT1, *OUT2;

/*******************************************************************************
                                 FUNCTION PROTOTYPES
*******************************************************************************/

/*** Top-Level Executive Functions ***/

  int InitializeArrays(void);
  int ReadInput(void);
  int GaussianElimination(void);
  int ComputeParameters(void);
  int WriteParameters(void);
  int CleanUp(void);

/*** Mid-Level Procedural Functions ***/  

  int Swaprow(int,int);
  int Eliminate(int,int);

/*** Mathematical and Computational Functions ***/


/*******************************************************************************
                               EXECUTIVE (MAIN) ROUTINE
*******************************************************************************/

  main(int argc, char * argv[])
{
  InitializeArrays();
  ReadInput();
  GaussianElimination();
  ComputeParameters();
  WriteParameters();
  CleanUp();  

  return(0);
}

/*******************************************************************************
  TOP-LEVEL:  INITIALIZE THE MATRIX AND ARRAYS
******************************************************************************/
  int InitializeArrays(void)
{
 int i, j;

 for (i=0; i<MATRIXDIMENSION; i++)
   {
    b[i] = zero;
    y[i] = zero;
    for (j=0; j<MATRIXDIMENSION; j++)
       {
        p[i][j] = zero;
       }
   }

 numvariables=NUMSPC + NUMEVENTS;

 return(0);
}

/*******************************************************************************
  TOP-LEVEL:  READ AND PROCESS INPUT
******************************************************************************/
  int ReadInput(void)
{
  double W, WTin;
  int retval, recordid;

  IN1=fopen("Hgdata.srt","r");
 
  while(!feof(IN1))
    {
     retval=fscanf(IN1,"%d %d %lf %lf %d %lf %d\n",
                   &SPCin, &Eventin, &Inchesin, &Resultin, &DLin, &WTin, &recordid);
     if (retval!=7) {break;}
   
     W = WTin;
/*   W = (double) WTin; */
     L = log(Inchesin + one); Y = log(onek*Resultin + one);
     p[SPCin][Eventin + NUMSPC]            += L * W;
     p[Eventin + NUMSPC][SPCin]            += L * W;
     p[SPCin][SPCin]                       += L * L * W;
     p[Eventin + NUMSPC][Eventin + NUMSPC] += one * W;
     y[SPCin]                              += L * Y * W;
     y[Eventin + NUMSPC]                   += Y * W;
    }

 fclose(IN1);

 return(0);
}

/******************************************************************************
  TOP-LEVEL:  GAUSSIAN ELIMINATION
******************************************************************************/
  int GaussianElimination(void)
{
 int i, k;

 for (i=1; i<numvariables; i++) /*Loop through all variables except the last.*/
    {
     printf("In Gaussian Elimination: working on variable %6d.\n", i);
     for (k=i+1; k<=numvariables; k++)
        {
         if (p[i][k] != zero) Eliminate(i,k);
        }
    }


 return(0);
}

/******************************************************************************
  TOP-LEVEL:  COMPUTE PARAMETERS   
******************************************************************************/
  int ComputeParameters(void)
{
 int i, j;

 b[numvariables] = y[numvariables] / p[numvariables][numvariables];

 for (i=numvariables-1; i>=1; i--)
   {
    for (j=i+1; j<=numvariables; j++)
       {
        y[i] -= b[j]*p[j][i];
       }
    b[i] = y[i] / p[i][i];
   }

 return(0);
}

/*******************************************************************************
  TOP-LEVEL:  WRITE PARAMETERS
******************************************************************************/
  int WriteParameters(void)
{
 int i;

 OUT1 = fopen("SPCparameters.txt", "w");
 for (i=1; i<=NUMSPC; i++)
   {
    fprintf(OUT1, "%4d  %12.7lf\n", i, b[i]);
   }
 fclose(OUT1);

 OUT2 = fopen("EVENTparameters.txt", "w");
 for (i=1; i<=NUMEVENTS; i++)
   {
    fprintf(OUT2, "%4d  %12.7lf\n", i, b[i+NUMSPC]);
   }
 fclose(OUT2);

 return(0);
}

/*******************************************************************************
  TOP-LEVEL:  CLEAN UP
******************************************************************************/
  int CleanUp(void)
{

 printf("Program terminates.\n");

 return(0);
}

/*******************************************************************************
  MID-LEVEL:  SWAP TWO ROWS IN THE MATRIX
******************************************************************************/

  /**** This function is not used in this version of this program. ****/

  int Swaprow(int high, int low)
{
 double temp;
 int i;

 for (i=1; i<=numvariables; i++)
    {
     temp = p[i][high];
     p[i][high] = p[i][low];
     p[i][low] = temp;
    }
 return(0);
}

/*******************************************************************************
  MID-LEVEL:  ELIMINATE A VARIABLE FROM A LOWER ROW IN THE MATRIX
******************************************************************************/
  int Eliminate(int var, int row)
{
 double A;
 int i;

 A = p[var][row] / p[var][var];

 for (i=var; i<=numvariables; i++)
    {
     p[i][row] -= A * p[i][var];
    }

    y[row] -= A * y[var];     
 
 return(0);
}
